/*
 * Copyright (c) 2016 Network New Technologies Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.networknt.db;

import com.networknt.config.Config;
import com.zaxxer.hikari.HikariDataSource;

import javax.sql.DataSource;
import java.util.Map;

/**
 * This is the generic Hikari datasource and it is suitable if you only have one database
 * in your application. You can still use it when you have one datasource for production
 * and another one for testing by putting two different version of service.yml and datasource.yml
 * in both main/resources/config and test/resources/config.
 *
 * @author Steve Hu
 */
public class GenericDataSource {
    protected static final String DATASOURCE = "datasource";
    protected static final String SECRET = "secret";
    private static final String DB_PASSWORD = "password";
    private static final String DS_NAME = "H2DataSource";

    // the HikariDataSource
    private HikariDataSource ds;
    // the data source name
    protected String dsName;
    protected Map<String, Object> dataSourceMap;

    public String getDsName() {
        return dsName;
    }

    public String getDbPassKey() {
        return DB_PASSWORD;
    }

    public GenericDataSource() {
        this.dsName = DS_NAME;
        this.ds = createDataSource();
    }

    public GenericDataSource(String dsName) {
        this.dsName = dsName;
        this.ds = createDataSource();
    }

    protected HikariDataSource createDataSource() {
        // get the configured datasources
        dataSourceMap = Config.getInstance().getJsonMapConfig(DATASOURCE);

        // get the decrypted secret file
        Map<String, Object> secret = Config.getInstance().getJsonMapConfig(SECRET);

        // get the requested datasource
        Map<String, Object> mainParams = (Map<String, Object>) dataSourceMap.get(getDsName());
        Map<String, String> configParams = (Map<String, String>)mainParams.get("parameters");

        // create the DataSource
        ds = new HikariDataSource();
        ds.setJdbcUrl((String)mainParams.get("jdbcUrl"));
        ds.setUsername((String)mainParams.get("username"));

        // use encrypted password
        String password = (String)mainParams.get(DB_PASSWORD);
        if (password==null) {
            password = (String)secret.get(getDbPassKey());
        }
        ds.setPassword(password);

        // set datasource paramters
        ds.setMaximumPoolSize((Integer)mainParams.get("maximumPoolSize"));
        ds.setConnectionTimeout((Integer)mainParams.get("connectionTimeout"));

        // add datasource specific connection parameters
        if(configParams != null) configParams.forEach((k, v) -> ds.addDataSourceProperty(k, v));

        return ds;
    }


    /**
     * Get an instance of the datasource
     *
     * @return the DataSource object
     */
    public HikariDataSource getDataSource() {
        return ds;
    }

}
